@@ -10,6 +10,7 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import datetime
 import itertools
 import json
 import logging
@@ -17,6 +18,7 @@ import sys
 import uuid
 
 import mock
+from oslo_utils import encodeutils
 import requests
 import requests.auth
 import six
@@ -322,7 +324,7 @@ class SessionTests(utils.TestCase):
         # Assert that response headers contains actual values and
         # only debug logs has been masked
         for k, v in security_headers.items():
-            self.assertIn('%s: {SHA1}' % k, self.logger.output)
+            self.assertIn('%s: {SHA256}' % k, self.logger.output)
             self.assertEqual(v, resp.headers[k])
             self.assertNotIn(v, self.logger.output)
 
@@ -438,6 +440,139 @@ class SessionTests(utils.TestCase):
         self.assertThat(self.requests_mock.request_history,
                         matchers.HasLength(retries + 1))
 
+    def test_connect_retries_interval_limit(self):
+        self.stub_url('GET', exc=requests.exceptions.Timeout())
+
+        session = client_session.Session()
+        retries = 20
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.ConnectTimeout,
+                              session.get,
+                              self.TEST_URL, connect_retries=retries)
+
+            self.assertEqual(retries, m.call_count)
+            # The interval maxes out at 60
+            m.assert_called_with(60.0)
+
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_connect_retries_fixed_delay(self):
+        self.stub_url('GET', exc=requests.exceptions.Timeout())
+
+        session = client_session.Session()
+        retries = 3
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.ConnectTimeout,
+                              session.get,
+                              self.TEST_URL, connect_retries=retries,
+                              connect_retry_delay=0.5)
+
+            self.assertEqual(retries, m.call_count)
+            m.assert_has_calls([mock.call(0.5)] * retries)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_http_503_retries(self):
+        self.stub_url('GET', status_code=503)
+
+        session = client_session.Session()
+        retries = 3
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.ServiceUnavailable,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries)
+
+            self.assertEqual(retries, m.call_count)
+            # 3 retries finishing with 2.0 means 0.5, 1.0 and 2.0
+            m.assert_called_with(2.0)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_http_status_retries(self):
+        self.stub_url('GET', status_code=409)
+
+        session = client_session.Session()
+        retries = 3
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.Conflict,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries,
+                              retriable_status_codes=[503, 409])
+
+            self.assertEqual(retries, m.call_count)
+            # 3 retries finishing with 2.0 means 0.5, 1.0 and 2.0
+            m.assert_called_with(2.0)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_http_status_retries_another_code(self):
+        self.stub_url('GET', status_code=404)
+
+        session = client_session.Session()
+        retries = 3
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.NotFound,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries,
+                              retriable_status_codes=[503, 409])
+
+            self.assertFalse(m.called)
+
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(1))
+
+    def test_http_status_retries_fixed_delay(self):
+        self.stub_url('GET', status_code=409)
+
+        session = client_session.Session()
+        retries = 3
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.Conflict,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries,
+                              status_code_retry_delay=0.5,
+                              retriable_status_codes=[503, 409])
+
+            self.assertEqual(retries, m.call_count)
+            m.assert_has_calls([mock.call(0.5)] * retries)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_http_status_retries_inverval_limit(self):
+        self.stub_url('GET', status_code=409)
+
+        session = client_session.Session()
+        retries = 20
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.Conflict,
+                              session.get,
+                              self.TEST_URL, status_code_retries=retries,
+                              retriable_status_codes=[503, 409])
+
+            self.assertEqual(retries, m.call_count)
+            # The interval maxes out at 60
+            m.assert_called_with(60.0)
+
+        # we count retries so there will be one initial request + 3 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
     def test_uses_tcp_keepalive_by_default(self):
         session = client_session.Session()
         requests_session = session.session
@@ -484,6 +619,118 @@ class SessionTests(utils.TestCase):
                    'value': 'new_name'}])
         self.assertContentTypeIs('application/json-patch+json')
 
+    def test_api_sig_error_message_single(self):
+        title = 'this error is bogus!'
+        detail = 'it is a totally made up error'
+        error_message = {
+            'errors': [
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': title,
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                }
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = '{} (HTTP 9000)'.format(title)
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertEqual(ex.details, detail)
+
+    def test_api_sig_error_message_multiple(self):
+        title = 'this error is the first error!'
+        detail = 'it is a totally made up error'
+        error_message = {
+            'errors': [
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': title,
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                },
+                {
+                    'request_id': uuid.uuid4().hex,
+                    'code': 'phoney.bologna.error',
+                    'status': 500,
+                    'title': 'some other error',
+                    'detail': detail,
+                    'links': [
+                        {
+                            'rel': 'help',
+                            'href': 'https://openstack.org'
+                        }
+                    ]
+                }
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = ('Multiple error responses, showing first only: {} (HTTP 9000)'
+               .format(title))
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertEqual(ex.details, detail)
+
+    def test_api_sig_error_message_empty(self):
+        error_message = {
+            'errors': [
+            ]
+        }
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        # The exception should contain the information from the error response
+        msg = 'HTTP Error (HTTP 9000)'
+
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+            self.assertIsNone(ex.details)
+
+    def test_error_message_unknown_schema(self):
+        error_message = 'Uh oh, things went bad!'
+        payload = json.dumps(error_message)
+        self.stub_url('GET', status_code=9000, text=payload,
+                      headers={'Content-Type': 'application/json'})
+        session = client_session.Session()
+
+        msg = 'Unrecognized schema in response body. (HTTP 9000)'
+        try:
+            session.get(self.TEST_URL)
+        except exceptions.HttpError as ex:
+            self.assertEqual(ex.message, msg)
+
 
 class RedirectTests(utils.TestCase):
 
@@ -956,6 +1203,86 @@ class SessionAuthTests(utils.TestCase):
         self.assertNotIn(list(response.keys())[0], self.logger.output)
         self.assertNotIn(list(response.values())[0], self.logger.output)
 
+    def test_split_loggers(self):
+
+        def get_logger_io(name):
+            logger_name = 'keystoneauth.session.{name}'.format(name=name)
+            logger = logging.getLogger(logger_name)
+            logger.setLevel(logging.DEBUG)
+
+            io = six.StringIO()
+            handler = logging.StreamHandler(io)
+            logger.addHandler(handler)
+            return io
+
+        io = {}
+        for name in ('request', 'body', 'response', 'request-id'):
+            io[name] = get_logger_io(name)
+
+        auth = AuthPlugin()
+        sess = client_session.Session(auth=auth, split_loggers=True)
+        response_key = uuid.uuid4().hex
+        response_val = uuid.uuid4().hex
+        response = {response_key: response_val}
+        request_id = uuid.uuid4().hex
+
+        self.stub_url(
+            'GET',
+            json=response,
+            headers={
+                'Content-Type': 'application/json',
+                'X-OpenStack-Request-ID': request_id,
+            })
+
+        resp = sess.get(
+            self.TEST_URL,
+            headers={
+                encodeutils.safe_encode('x-bytes-header'):
+                encodeutils.safe_encode('bytes-value')
+            })
+
+        self.assertEqual(response, resp.json())
+
+        request_output = io['request'].getvalue().strip()
+        response_output = io['response'].getvalue().strip()
+        body_output = io['body'].getvalue().strip()
+        id_output = io['request-id'].getvalue().strip()
+
+        self.assertIn('curl -g -i -X GET {url}'.format(url=self.TEST_URL),
+                      request_output)
+        self.assertIn('-H "x-bytes-header: bytes-value"', request_output)
+        self.assertEqual('[200] Content-Type: application/json '
+                         'X-OpenStack-Request-ID: '
+                         '{id}'.format(id=request_id), response_output)
+        self.assertEqual(
+            'GET call to {url} used request id {id}'.format(
+                url=self.TEST_URL, id=request_id),
+            id_output)
+        self.assertEqual(
+            '{{"{key}": "{val}"}}'.format(
+                key=response_key, val=response_val),
+            body_output)
+
+    def test_collect_timing(self):
+        auth = AuthPlugin()
+        sess = client_session.Session(auth=auth, collect_timing=True)
+        response = {uuid.uuid4().hex: uuid.uuid4().hex}
+
+        self.stub_url('GET',
+                      json=response,
+                      headers={'Content-Type': 'application/json'})
+
+        resp = sess.get(self.TEST_URL)
+
+        self.assertEqual(response, resp.json())
+        timings = sess.get_timings()
+        self.assertEqual(timings[0].method, 'GET')
+        self.assertEqual(timings[0].url, self.TEST_URL)
+        self.assertIsInstance(timings[0].elapsed, datetime.timedelta)
+        sess.reset_timings()
+        timings = sess.get_timings()
+        self.assertEqual(len(timings), 0)
+
 
 class AdapterTest(utils.TestCase):
 
@@ -1008,19 +1335,25 @@ class AdapterTest(utils.TestCase):
         self.assertRequestHeaderEqual('User-Agent', self.USER_AGENT)
 
     def test_setting_global_id_on_request(self):
-        global_id = "req-%s" % uuid.uuid4()
+        global_id_adpt = "req-%s" % uuid.uuid4()
+        global_id_req = "req-%s" % uuid.uuid4()
         response = uuid.uuid4().hex
         self.stub_url('GET', text=response)
-        adpt = adapter.Adapter(client_session.Session(),
-                               auth=CalledAuthPlugin(),
-                               service_type=self.SERVICE_TYPE,
-                               service_name=self.SERVICE_NAME,
-                               interface=self.INTERFACE,
-                               region_name=self.REGION_NAME,
-                               user_agent=self.USER_AGENT,
-                               version=self.VERSION,
-                               allow=self.ALLOW,
-                               global_request_id=global_id)
+
+        def mk_adpt(**kwargs):
+            return adapter.Adapter(client_session.Session(),
+                                   auth=CalledAuthPlugin(),
+                                   service_type=self.SERVICE_TYPE,
+                                   service_name=self.SERVICE_NAME,
+                                   interface=self.INTERFACE,
+                                   region_name=self.REGION_NAME,
+                                   user_agent=self.USER_AGENT,
+                                   version=self.VERSION,
+                                   allow=self.ALLOW,
+                                   **kwargs)
+
+        # No global_request_id
+        adpt = mk_adpt()
         resp = adpt.get('/')
         self.assertEqual(resp.text, response)
 
@@ -1028,7 +1361,21 @@ class AdapterTest(utils.TestCase):
         self.assertEqual(self.ALLOW,
                          adpt.auth.endpoint_arguments['allow'])
         self.assertTrue(adpt.auth.get_token_called)
-        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', None)
+
+        # global_request_id only on the request
+        adpt.get('/', global_request_id=global_id_req)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_req)
+
+        # global_request_id only on the adapter
+        adpt = mk_adpt(global_request_id=global_id_adpt)
+        adpt.get('/')
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_adpt)
+
+        # global_request_id on the adapter *and* the request (the request takes
+        # precedence)
+        adpt.get('/', global_request_id=global_id_req)
+        self.assertRequestHeaderEqual('X-OpenStack-Request-ID', global_id_req)
 
     def test_setting_variables_on_get_endpoint(self):
         adpt = self._create_loaded_adapter()
@@ -1135,6 +1482,39 @@ class AdapterTest(utils.TestCase):
         self.assertThat(self.requests_mock.request_history,
                         matchers.HasLength(retries + 1))
 
+    def test_adapter_http_503_retries(self):
+        retries = 2
+        sess = client_session.Session()
+        adpt = adapter.Adapter(sess, status_code_retries=retries)
+
+        self.stub_url('GET', status_code=503)
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.ServiceUnavailable,
+                              adpt.get, self.TEST_URL)
+            self.assertEqual(retries, m.call_count)
+
+        # we count retries so there will be one initial request + 2 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
+    def test_adapter_http_status_retries(self):
+        retries = 2
+        sess = client_session.Session()
+        adpt = adapter.Adapter(sess, status_code_retries=retries,
+                               retriable_status_codes=[503, 409])
+
+        self.stub_url('GET', status_code=409)
+
+        with mock.patch('time.sleep') as m:
+            self.assertRaises(exceptions.Conflict,
+                              adpt.get, self.TEST_URL)
+            self.assertEqual(retries, m.call_count)
+
+        # we count retries so there will be one initial request + 2 retries
+        self.assertThat(self.requests_mock.request_history,
+                        matchers.HasLength(retries + 1))
+
     def test_user_and_project_id(self):
         auth = AuthPlugin()
         sess = client_session.Session()
@@ -1307,7 +1687,7 @@ class AdapterTest(utils.TestCase):
 
         adap.get(url)
 
-        agent = 'ksatest/1.2.3 one/1.1.1 two/2.2.2 testclient/4.5.6'
+        agent = 'ksatest/1.2.3 testclient/4.5.6 one/1.1.1 two/2.2.2'
         self.assertEqual(agent + ' ' + client_session.DEFAULT_USER_AGENT,
                          self.requests_mock.last_request.headers['User-Agent'])
 
@@ -1394,6 +1774,7 @@ class AdapterTest(utils.TestCase):
             with mock.patch.object(sess, 'request') as m:
                 adapter.Adapter(sess, **adap_kwargs).get(url, **get_kwargs)
                 m.assert_called_once_with(url, 'GET', endpoint_filter={},
+                                          rate_semaphore=mock.ANY,
                                           **exp_kwargs)
 
         # No default_microversion in Adapter, no microversion in get()
@@ -1409,6 +1790,34 @@ class AdapterTest(utils.TestCase):
         validate({'default_microversion': '1.2'}, {'microversion': '1.5'},
                  {'microversion': '1.5'})
 
+    def test_raise_exc_override(self):
+        sess = client_session.Session()
+        url = 'http://url'
+
+        def validate(adap_kwargs, get_kwargs, exp_kwargs):
+            with mock.patch.object(sess, 'request') as m:
+                adapter.Adapter(sess, **adap_kwargs).get(url, **get_kwargs)
+                m.assert_called_once_with(url, 'GET', endpoint_filter={},
+                                          rate_semaphore=mock.ANY,
+                                          **exp_kwargs)
+
+        # No raise_exc in Adapter or get()
+        validate({}, {}, {})
+
+        # Set in Adapter, unset in get()
+        validate({'raise_exc': True}, {}, {'raise_exc': True})
+        validate({'raise_exc': False}, {}, {'raise_exc': False})
+
+        # Unset in Adapter, set in get()
+        validate({}, {'raise_exc': True}, {'raise_exc': True})
+        validate({}, {'raise_exc': False}, {'raise_exc': False})
+
+        # Setting in get() overrides the one in Adapter
+        validate({'raise_exc': True}, {'raise_exc': False},
+                 {'raise_exc': False})
+        validate({'raise_exc': False}, {'raise_exc': True},
+                 {'raise_exc': True})
+
 
 class TCPKeepAliveAdapterTest(utils.TestCase):
 
